/* Automatically generated by refactor.pl.
 *
 *
 * CMDNAME: banlist
 * CMDLEVEL: QCMD_AUTHED
 * CMDARGS: 1
 * CMDDESC: Displays all persistent bans on a channel.
 * CMDFUNC: csc_dobanlist
 * CMDPROTO: int csc_dobanlist(void *source, int cargc, char **cargv);
 * CMDHELP: Usage: BANLIST <channel>
 * CMDHELP: Displays a list of persistent and channel bans on the named channel.  Each ban
 * CMDHELP: is identified by a number which can be passed to a subsequent BANDEL command.
 * CMDHELP: Note that the numbers can change if bans are added to or removed from the 
 * CMDHELP: channel. Where:
 * CMDHELP: channel - the channel to use
 * CMDHELP: BANLIST requires operator (+o) access on the named channel.
 */

#include "../chanserv.h"
#include "../../nick/nick.h"
#include "../../lib/flags.h"
#include "../../lib/irc_string.h"
#include "../../channel/channel.h"
#include "../../parser/parser.h"
#include "../../irc/irc.h"
#include "../../localuser/localuserchannel.h"
#include <string.h>
#include <stdio.h>

int csc_dobanlist(void *source, int cargc, char **cargv) {
  nick *sender=source;
  chanindex *cip;
  regchan *rcp;
  regban *rbp;
  reguser *rup;
  chanban *cbp;
  int i=0;
  int exp;

  if (cargc<1) {
    chanservstdmessage(sender, QM_NOTENOUGHPARAMS, "banlist");
    return CMD_ERROR;
  }
  
  if(!(cip=cs_checkaccess(sender, cargv[0], CA_OPPRIV, NULL, "banlist", 0, 0)))
    return CMD_ERROR;

  rcp=cip->exts[chanservext];
  
  if (rcp->bans || (cip->channel && cip->channel->bans)) {
    chanservstdmessage(sender, QM_REGBANHEADER, cip->name->content);
    for(rbp=rcp->bans;rbp;rbp=rbp->next) {
      rup=findreguserbyID(rbp->setby);
      chanservsendmessage(sender," #%-2d %-29s %-18s %-15s  %s",++i,bantostring(rbp->cbp),
			  rbp->expiry?longtoduration(rbp->expiry-time(NULL),0):"Permanent",
			  rup?rup->username:"<unknown>",
			  rbp->reason?rbp->reason->content:"");
    }
    if (cip->channel) {
      for (cbp=cip->channel->bans;cbp;cbp=cbp->next) {
        for (rbp=rcp->bans;rbp;rbp=rbp->next) {
          if (banequal(rbp->cbp, cbp))
	    break;
        }
        if (!rbp) {
	  if (rcp->banduration) {
	    exp=(cbp->timeset + rcp->banduration) - time(NULL);
          } else {
	    exp=0;
          } 
          chanservsendmessage(sender, " #%-2d %-29s %-18s %-15s",++i,bantostring(cbp),
	    		      exp ? longtoduration(exp,0) : "Permanent",
	    		      "(channel ban)");
        }
      }
    }
    chanservstdmessage(sender, QM_ENDOFLIST);
  } else {
    chanservstdmessage(sender, QM_NOBANS, cip->name->content);
  }
      
  return CMD_OK;
}
